module Michelson.TypeCheck.Value
    ( typeCheckValImpl
    , typeCheckCValue
    ) where

import Control.Monad.Except (liftEither, throwError)
import Data.Constraint (Dict(..))
import Data.Default (def)
import qualified Data.Map as M
import qualified Data.Set as S
import Data.Singletons (SingI)
import Data.Typeable ((:~:)(..))
import Fmt (pretty)
import Prelude hiding (EQ, GT, LT)

import Michelson.Text
import Michelson.TypeCheck.Error (TCError(..), TCTypeError(..))
import Michelson.TypeCheck.Helpers
import Michelson.TypeCheck.TypeCheck (TcInstrHandler, TypeCheckEnv(..), TypeCheckInstr)
import Michelson.TypeCheck.Types
import Michelson.Typed
  (CT(..), CValue(..), Notes(..), Sing(..), Value'(..), converge, fromSingCT, fromSingT, starNotes)
import qualified Michelson.Typed as T
import qualified Michelson.Untyped as U
import Tezos.Address (Address(..), parseAddress)
import Tezos.Core
import Tezos.Crypto (parseKeyHash, parsePublicKey, parseSignature)

typeCheckCValue
  :: U.Value' op -> CT -> Either (U.Value' op, TCTypeError) SomeCValue
typeCheckCValue (U.ValueInt i) CInt = pure $ CvInt i :--: SCInt
typeCheckCValue (U.ValueInt i) CNat
  | i >= 0 = pure $ CvNat (fromInteger i) :--: SCNat
  | otherwise = Left (U.ValueInt i, NegativeNat)
typeCheckCValue v@(U.ValueInt i) CMutez = do
  mtz <- maybeToRight (v, InvalidTimestamp) . mkMutez $ fromInteger i
  pure $ CvMutez mtz :--: SCMutez
typeCheckCValue (U.ValueString s) CString =
  pure $ CvString s :--: SCString
typeCheckCValue v@(U.ValueString s) CAddress = do
  addr <- T.parseEpAddress (unMText s) `onLeft` ((v, ) . InvalidAddress)
  pure $ CvAddress addr :--: SCAddress
typeCheckCValue v@(U.ValueString s) CKeyHash = do
  kHash <- parseKeyHash (unMText s) `onLeft` ((v, ) . InvalidKeyHash)
  pure $ CvKeyHash kHash :--: SCKeyHash
typeCheckCValue v@(U.ValueString s) CTimestamp = do
  tstamp <- maybeToRight (v, InvalidTimestamp) . parseTimestamp $ unMText s
  pure $ CvTimestamp tstamp :--: SCTimestamp
typeCheckCValue (U.ValueInt i) CTimestamp =
  pure $ CvTimestamp (timestampFromSeconds i) :--: SCTimestamp
typeCheckCValue (U.ValueBytes (U.InternalByteString s)) CBytes =
  pure $ CvBytes s :--: SCBytes
typeCheckCValue U.ValueTrue CBool = pure $ CvBool True :--: SCBool
typeCheckCValue U.ValueFalse CBool = pure $ CvBool False :--: SCBool
typeCheckCValue v t =
  Left $ (v, InvalidValueType (T.Tc t))

typeCheckCVals
  :: forall t op . (Typeable t, SingI t)
  => [U.Value' op]
  -> CT
  -> Either (U.Value' op, TCTypeError) [CValue t]
typeCheckCVals mvs t = traverse check mvs
  where
    check mv = do
      v :--: (_ :: Sing t') <- typeCheckCValue mv t
      Refl <- eqType @('T.Tc t) @('T.Tc t') `onLeft` (,) mv
      pure v

tcFailedOnValue :: U.Value -> T.T -> Text -> Maybe TCTypeError -> TypeCheckInstr a
tcFailedOnValue v t msg err = do
  loc <- ask
  throwError $ TCFailedOnValue v t msg loc err

-- | Function @typeCheckValImpl@ converts a single Michelson value
-- given in representation from @Michelson.Type@ module to representation
-- in strictly typed GADT.
--
-- As a third argument, @typeCheckValImpl@ accepts expected type of value.
--
-- Type checking algorithm pattern-matches on parse value representation,
-- expected type @t@ and constructs @Val t@ value.
--
-- If there was no match on a given pair of value and expected type,
-- that is interpreted as input of wrong type and type check finishes with
-- error.
typeCheckValImpl
  :: TcInstrHandler
  -> U.Value
  -> (Sing t, Notes t)
  -> TypeCheckInstr SomeNotedValue
typeCheckValImpl _ mv (t@(STc ct), NTc nt) = do
  case typeCheckCValue mv (fromSingCT ct) of
    Left (uval, err) -> tcFailedOnValue uval (fromSingT $ t) "" (Just err)
    Right (v :--: cst) -> pure $ VC v :::: (STc cst, NTc nt)
typeCheckValImpl _ (U.ValueString (parsePublicKey . unMText -> Right s)) t@(STKey, _) =
  pure $ T.VKey s :::: t

typeCheckValImpl _ (U.ValueString (parseSignature . unMText -> Right s)) t@(STSignature, _) =
  pure $ VSignature s :::: t

typeCheckValImpl _ (U.ValueString (parseChainId . unMText -> Right ci)) t@(STChainId, _) =
  pure $ VChainId ci :::: t
typeCheckValImpl _ (U.ValueBytes (mkChainId . U.unInternalByteString -> Just ci)) t@(STChainId, _) =
  pure $ VChainId ci :::: t

typeCheckValImpl _ cv@(U.ValueString (parseAddress . unMText -> Right addr))
                   t@(STContract pc, cn) = do
  instrPos <- ask
  let tcFail = \msg -> TCFailedOnValue cv (fromSingT $ fst t) msg instrPos Nothing
  let NTContract _ pn = cn
  contracts <- gets tcContracts
  let ensureTypeMatches cp =
        liftEither $
        first (TCFailedOnValue cv (fromSingT $ fst t) "invalid contract parameter" instrPos . Just) $
          compareTypes (pc, pn) cp
  -- TODO [TM-280]: Actually the desired entrypoint can be specified in
  -- "<address>%<entrypoint>" format. Support this scenario.
  let unsupportedType :: Text -> Either TCError a
      unsupportedType desc =
        Left $
        TCFailedOnValue cv (fromSingT pc) (desc <> " in type argument of 'contract' type") instrPos $
        Just (UnsupportedTypes [fromSingT pc])
  Dict <- liftEither $ maybe (unsupportedType "Operation") pure (T.opAbsense pc)
  Dict <- liftEither $ maybe (unsupportedType "Nested BigMaps") pure (T.nestedBigMapsAbsense pc)
  let entrypoint = def
  case addr of
    KeyAddress _ -> do
      ensureTypeMatches U.tyImplicitAccountParam
      pure $ VContract addr entrypoint :::: t
    ContractAddress ca ->
      case M.lookup ca contracts of
        Just contractParam -> do
          ensureTypeMatches contractParam
          pure $ VContract addr entrypoint :::: t
        _ -> throwError $ tcFail $ "Contract literal " <> pretty addr <> " doesn't exist"

typeCheckValImpl _ U.ValueUnit t@(STUnit, _) = pure $ VUnit :::: t
typeCheckValImpl
  tcDo
  (U.ValuePair ml mr)
  (STPair lt rt, NTPair n1 n2 n3 nl nr)
    = do
  l :::: (lst, ln) <- typeCheckValImpl tcDo ml (lt, nl)
  r :::: (rst, rn) <- typeCheckValImpl tcDo mr (rt, nr)
  pure $ VPair (l, r) :::: (STPair lst rst, NTPair n1 n2 n3 ln rn)
typeCheckValImpl tcDo (U.ValueLeft ml) (STOr lt rt, NTOr n1 n2 n3 nl nr) = do
  l :::: (lst, ln) <- typeCheckValImpl tcDo ml (lt, nl)
  pure $ VOr (Left l) :::: ( STOr lst rt, NTOr n1 n2 n3 ln nr)
typeCheckValImpl tcDo (U.ValueRight mr) (STOr lt rt, NTOr n1 n2 n3 nl nr) = do
  r :::: (rst, rn) <- typeCheckValImpl tcDo mr (rt, nr)
  pure $ VOr (Right r) :::: ( STOr lt rst, NTOr n1 n2 n3 nl rn)
typeCheckValImpl tcDo (U.ValueSome mv) (STOption vt, NTOption na nt) = do
  v :::: (vst, vns) <- typeCheckValImpl tcDo mv (vt, nt)
  pure $ VOption (Just v) :::: (STOption vst, NTOption na vns)
typeCheckValImpl _ U.ValueNone t@(STOption _, _) =
  pure $ VOption Nothing :::: t

typeCheckValImpl _ U.ValueNil t@(STList _, _) =
  pure $ T.VList [] :::: t

typeCheckValImpl
  tcDo
  (U.ValueSeq (toList -> mels))
  t@(STList vt, NTList _ x)
    = do
  (els, _) <- typeCheckValsImpl tcDo mels (vt, x)
  pure $ VList els :::: t

typeCheckValImpl _ U.ValueNil t@(STSet _, _) = pure $ T.VSet S.empty :::: t

typeCheckValImpl _ sq@(U.ValueSeq (toList -> mels)) t@(STSet vt, _) = do
  instrPos <- ask
  els <- liftEither $ typeCheckCVals mels (fromSingCT vt)
          `onLeft` \(cv, err) -> TCFailedOnValue cv (fromSingT $ STc vt)
                                      "wrong type of set element:" instrPos (Just err)
  elsS <- liftEither $ S.fromDistinctAscList <$> ensureDistinctAsc id els
            `onLeft` \msg -> TCFailedOnValue sq (fromSingT $ STc vt) msg instrPos Nothing
  pure $ VSet elsS :::: t

typeCheckValImpl _ U.ValueNil t@(STMap _ _, _) = pure $ T.VMap M.empty :::: t

typeCheckValImpl
  tcDo
  sq@(U.ValueMap (toList -> mels))
  t@(STMap kt vt, NTMap _ _ vn)
    = do
  keyOrderedElts <- typeCheckMapVal tcDo mels sq vn kt vt
  pure $ VMap (M.fromDistinctAscList keyOrderedElts) :::: t

typeCheckValImpl _ U.ValueNil t@(STBigMap _ _ , _) = pure $ T.VBigMap M.empty :::: t

typeCheckValImpl
  tcDo
  sq@(U.ValueMap (toList -> mels))
  t@(STBigMap kt vt, NTBigMap _ _ vn)
    = do
  keyOrderedElts <- typeCheckMapVal tcDo mels sq vn kt vt
  pure $ VBigMap (M.fromDistinctAscList keyOrderedElts) :::: t

typeCheckValImpl
  tcDo
  v
  ( t@(STLambda (it :: Sing it) (ot :: Sing ot))
  , NTLambda vn _ _
  )
    = do
  mp <- case v of
    U.ValueNil       -> pure []
    U.ValueLambda mp -> pure $ toList mp
    _ -> tcFailedOnValue v (fromSingT t) "unexpected value" Nothing
  li :/ instr <- typeCheckImpl tcDo mp ((it, starNotes, def) ::& SNil)
  let (_, ins, _) ::& SNil = li
  let lamS = STLambda it ot
  case instr of
    lam ::: (lo :: HST lo) -> do
      case eqHST1 @ot lo of
        Right Refl -> do
          let (_, ons, _) ::& SNil = lo
          pure $ VLam (T.RfNormal lam) :::: (STLambda it ot, NTLambda vn ins ons)
        Left m ->
          tcFailedOnValue v (fromSingT t)
                  "wrong output type of lambda's value:" (Just m)
    AnyOutInstr lam ->
      pure $ VLam (T.RfAlwaysFails lam) :::: (lamS, NTLambda def ins starNotes)

typeCheckValImpl _ v (t, _) = tcFailedOnValue v (fromSingT t) "unknown value" Nothing

-- | Function @typeCheckMapVal@ typechecks given list of @Elt@s and
-- ensures, that its keys are in ascending order.
--
-- It return list of pairs (key, value) with keys in ascending order
-- so it is safe to call @fromDistinctAscList@ on returned list
typeCheckMapVal
  :: (SingI kt, Typeable kt, SingI vt, Typeable vt)
  => TcInstrHandler
  -> [U.Elt U.ExpandedOp]
  -> U.Value
  -> Notes vt
  -> Sing kt
  -> Sing vt
  -> TypeCheckInstr [(CValue kt, T.Value vt)]
typeCheckMapVal tcDo mels sq vn kt vt = do
  instrPos <- ask
  ks <- liftEither $ typeCheckCVals (map (\(U.Elt k _) -> k) mels) (fromSingCT kt)
          `onLeft` \(cv, err) -> TCFailedOnValue cv (fromSingT $ STc kt)
                                      "wrong type of map key:" instrPos (Just err)
  (vals, _) <- typeCheckValsImpl tcDo (map (\(U.Elt _ v) -> v) mels) (vt, vn)
  ksS <- liftEither $ ensureDistinctAsc id ks
        `onLeft` \msg -> TCFailedOnValue sq (fromSingT $ STc kt) msg instrPos Nothing
  pure $ zip ksS vals

typeCheckValsImpl
  :: forall t . (Typeable t, SingI t)
  => TcInstrHandler
  -> [U.Value]
  -> (Sing t, Notes t)
  -> TypeCheckInstr ([T.Value t], Notes t)
typeCheckValsImpl tcDo mvs (t, nt) =
  fmap (first reverse) $ foldM check ([], nt) mvs
  where
    check (res, ns) mv = do
      instrPos <- ask
      v :::: ((_ :: Sing t'), vns) <- typeCheckValImpl tcDo mv (t, nt)
      Refl <- liftEither $ eqType @t @t' `onLeft`
        (TCFailedOnValue mv (fromSingT t) "wrong element type" instrPos . Just)
      ns' <- liftEither $ converge ns vns `onLeft`
        ((TCFailedOnValue mv (fromSingT t) "wrong element type") instrPos . Just . AnnError)
      pure (v : res, ns')
